# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Fix compatibility with werkzeug due to flask_oauthlib imports."""

import sys

import werkzeug


def monkey_patch_werkzeug_contrib():
    """Patch missing module."""
    try:
        from werkzeug import contrib
    except ImportError:
        import cachelib

        sys.modules["werkzeug.contrib.cache"] = cachelib


def monkey_patch_werkzeug_base():
    """Patch top level removed modules."""
    try:
        from werkzeug import cached_property
    except ImportError:
        werkzeug.cached_property = werkzeug.utils.cached_property
        werkzeug.parse_options_header = werkzeug.http.parse_options_header
        werkzeug.url_quote = werkzeug.urls.url_quote
        werkzeug.url_decode = werkzeug.urls.url_decode
        werkzeug.url_encode = werkzeug.urls.url_encode


def monkey_patch_werkzeug():
    """Patch all the missing werkzeug modules."""
    monkey_patch_werkzeug_base()
    monkey_patch_werkzeug_contrib()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import shutil
import tempfile
from unittest.mock import MagicMock

import pytest
from flask import Flask, url_for
from flask.cli import ScriptInfo
from flask.views import MethodView
from flask_babelex import Babel
from flask_breadcrumbs import Breadcrumbs
from flask_mail import Mail
from flask_menu import Menu
from helpers import create_oauth_client, patch_request
from invenio_accounts import InvenioAccountsREST, InvenioAccountsUI
from invenio_accounts.models import User
from invenio_accounts.views import blueprint as accounts_blueprint
from invenio_db import InvenioDB, db
from six import get_method_self
from sqlalchemy_utils.functions import create_database, database_exists, drop_database

from invenio_oauth2server import InvenioOAuth2Server, InvenioOAuth2ServerREST
from invenio_oauth2server.decorators import require_api_auth, require_oauth_scopes
from invenio_oauth2server.models import Client, Scope, Token
from invenio_oauth2server.views import server_blueprint, settings_blueprint

import invenio_oauth2server._compat  # noqa isort:skip

try:
    from werkzeug.middleware.dispatcher import DispatcherMiddleware
except ImportError:
    from werkzeug.wsgi import DispatcherMiddleware


@pytest.fixture()
def app(request):
    """Flask application fixture."""
    instance_path = tempfile.mkdtemp()

    def init_app(app):
        app.config.update(
            DB_VERSIONING=True,
            LOGIN_DISABLED=False,
            MAIL_SUPPRESS_SEND=True,
            OAUTH2_CACHE_TYPE="simple",
            OAUTHLIB_INSECURE_TRANSPORT=True,
            SECRET_KEY="CHANGE_ME",
            SECURITY_DEPRECATED_PASSWORD_SCHEMES=[],
            SECURITY_PASSWORD_HASH="plaintext",
            SECURITY_PASSWORD_SINGLE_HASH=["plaintext"],
            SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
            SECURITY_PASSWORD_SCHEMES=["plaintext"],
            SQLALCHEMY_DATABASE_URI=os.getenv(
                "SQLALCHEMY_DATABASE_URI",
                "sqlite:///" + os.path.join(instance_path, "test.db"),
            ),
            SQLALCHEMY_TRACK_MODIFICATIONS=True,
            TESTING=True,
            WTF_CSRF_ENABLED=False,
            APP_THEME="semantic-ui",
            THEME_ICONS={
                "semantic-ui": {
                    "key": "key icon",
                    "link": "linkify icon",
                    "shield": "shield alternate icon",
                    "user": "user icon",
                    "codepen": "codepen icon",
                    "cogs": "cogs icon",
                    # Special catch all:
                    "*": "{} icon",
                }
            },
            ACCOUNTS_COVER_TEMPLATE="invenio_accounts/base_cover.html",
            ACCOUNTS_BASE_TEMPLATE="invenio_accounts/base.html",
        )
        Babel(app)
        Mail(app)
        Menu(app)
        Breadcrumbs(app)
        InvenioDB(app)
        InvenioOAuth2Server(app)

    api_app = Flask("testapiapp", instance_path=instance_path)
    api_app.config.update(APPLICATION_ROOT="/api", ACCOUNTS_REGISTER_BLUEPRINT=True)
    init_app(api_app)
    InvenioAccountsREST(api_app)
    InvenioOAuth2ServerREST(api_app)

    app = Flask("testapp", instance_path=instance_path)
    init_app(app)
    InvenioAccountsUI(app)
    app.register_blueprint(accounts_blueprint)
    app.register_blueprint(server_blueprint)
    app.register_blueprint(settings_blueprint)

    app.wsgi_app = DispatcherMiddleware(app.wsgi_app, {"/api": api_app.wsgi_app})

    with app.app_context():
        if str(db.engine.url) != "sqlite://" and not database_exists(
            str(db.engine.url)
        ):
            create_database(str(db.engine.url))
        db.create_all()

    def teardown():
        with app.app_context():
            if str(db.engine.url) != "sqlite://":
                drop_database(str(db.engine.url))
            shutil.rmtree(instance_path)

    request.addfinalizer(teardown)
    return app


@pytest.fixture()
def api_app(app):
    """Retrieve the REST API application."""
    return get_method_self(app.wsgi_app.mounts["/api"])


@pytest.fixture()
def api_app_with_test_view(api_app):
    """Add a test view URL."""
    api_app.add_url_rule("/test", "test", view_func=lambda: "OK")
    return api_app


@pytest.fixture
def script_info(app):
    """Get ScriptInfo object for testing CLI."""
    return ScriptInfo(create_app=lambda: app)


@pytest.fixture
def settings_fixture(app):
    """Fixture for testing settings views."""
    from invenio_oauth2server.proxies import current_oauth2server

    with app.app_context():
        with db.session.begin_nested():
            datastore = app.extensions["security"].datastore
            datastore.create_user(email="info@inveniosoftware.org", password="tester")
        db.session.commit()
        current_oauth2server.register_scope(Scope("test:scope"))
        current_oauth2server.register_scope(Scope("test:scope2"))
    return app


@pytest.fixture
def developer_app_fixture(settings_fixture):
    """Fixture for testing developer application use cases."""
    settings_app = settings_fixture
    with settings_app.app_context():
        with db.session.begin_nested():
            datastore = settings_app.extensions["security"].datastore
            dev_user = datastore.create_user(
                email="dev@inveniosoftware.org", password="dev", active=True
            )

            dev_client = Client(
                client_id="dev",
                client_secret="dev",
                name="Test name",
                description="Test description",
                is_confidential=False,
                user=dev_user,
                website="http://inveniosoftware.org",
                _redirect_uris="",
                _default_scopes="test:scope",
            )

            user = datastore.get_user("info@inveniosoftware.org")
            user_token = Token(
                client=dev_client,
                user=user,
                token_type="bearer",
                access_token="dev_access_1",
                refresh_token="dev_refresh_1",
                expires=None,
                is_personal=False,
                is_internal=False,
                _scopes="test:scope",
            )

            db.session.add(dev_client)
            db.session.add(user_token)
        db.session.commit()
    return settings_app


@pytest.fixture
def models_fixture(app):
    """Fixture that contains the test data for models tests."""
    from invenio_oauth2server.proxies import current_oauth2server

    with app.app_context():
        # Register a test scope
        current_oauth2server.register_scope(Scope("test:scope1"))
        current_oauth2server.register_scope(Scope("test:scope2", internal=True))
        datastore = app.extensions["security"].datastore
        with db.session.begin_nested():
            test_user = datastore.create_user(
                email="info@inveniosoftware.org",
                password="tester",
            )
            resource_owner = datastore.create_user(
                email="resource_owner@inveniosoftware.org", password="test"
            )
            consumer = datastore.create_user(
                email="consumer@inveniosoftware.org", password="test"
            )

            # create resource_owner -> client_1
            u1c1 = Client(
                client_id="client_test_u1c1",
                client_secret="client_test_u1c1",
                name="client_test_u1c1",
                description="",
                is_confidential=False,
                user=resource_owner,
                _redirect_uris="",
                _default_scopes="",
            )
            # create resource_owner -> client_1 / resource_owner -> token_1

            u1c1u1t1 = Token(
                client=u1c1,
                user=resource_owner,
                token_type="u",
                access_token="dev_access_1",
                refresh_token="dev_refresh_1",
                expires=None,
                is_personal=False,
                is_internal=False,
                _scopes="",
            )

            # create consumer -> client_1 / resource_owner -> token_2

            u1c1u2t2 = Token(
                client=u1c1,
                user=consumer,
                token_type="u",
                access_token="dev_access_2",
                refresh_token="dev_refresh_2",
                expires=None,
                is_personal=False,
                is_internal=False,
                _scopes="",
            )
            db.session.add(u1c1)
            db.session.add(u1c1u1t1)
            db.session.add(u1c1u2t2)
        db.session.commit()
        test_user_id = test_user.get_id()
        app.test_user = lambda: User.query.get(test_user_id)
        app.resource_owner_id = resource_owner.get_id()
        app.consumer_id = consumer.get_id()
        app.u1c1_id = u1c1.client_id
        app.u1c1u1t1_id = u1c1u1t1.id
        app.u1c1u2t2_id = u1c1u2t2.id
    return app


@pytest.fixture
def provider_fixture(app):
    """Fixture that contains test data for provider tests."""
    from invenio_oauth2server.proxies import current_oauth2server

    # Mock the oauth client calls to prevent them from going online.
    oauth_client = create_oauth_client(app, "oauth2test")
    oauth_client.http_request = MagicMock(side_effect=patch_request(app))
    datastore = app.extensions["security"].datastore
    with app.test_request_context():
        with db.session.begin_nested():
            current_oauth2server.register_scope(Scope("test:scope"))

            user1 = datastore.create_user(
                email="info@inveniosoftware.org",
                password="tester",
                active=True,
            )
            datastore.create_user(
                email="abuse@inveniosoftware.org", password="tester2", active=True
            )
            user3 = datastore.create_user(
                email="inactive@inveniosoftware.org", password="tester3", active=False
            )

            c1 = Client(
                client_id="dev",
                client_secret="dev",
                name="dev",
                description="",
                is_confidential=False,
                user=user1,
                _redirect_uris=url_for("oauth2test.authorized", _external=True),
                _default_scopes="test:scope",
            )
            c2 = Client(
                client_id="confidential",
                client_secret="confidential",
                name="confidential",
                description="",
                is_confidential=True,
                user=user1,
                _redirect_uris=url_for("oauth2test.authorized", _external=True),
                _default_scopes="test:scope",
            )
            # Same as 'c2' but user belonging to a user that's inactive
            c3 = Client(
                client_id="confidential-user-inactive",
                client_secret="confidential-user-inactive",
                name="confidential-user-inactive",
                description="",
                is_confidential=True,
                user=user3,
                _redirect_uris=url_for("oauth2test.authorized", _external=True),
                _default_scopes="test:scope",
            )
            c4 = Client(
                client_id="confidential-email",
                client_secret="confidential-email",
                name="confidential-email",
                description="",
                is_confidential=True,
                user=user1,
                _redirect_uris=url_for("oauth2test.authorized", _external=True),
                _default_scopes="email",
            )
            db.session.add(c1)
            db.session.add(c2)
            db.session.add(c3)
            db.session.add(c4)
        personal_token = Token.create_personal(
            "test-personal", user1.id, scopes=[], is_internal=True
        )

        personal_token3 = Token.create_personal(
            "test-personal", user3.id, scopes=[], is_internal=True
        )
        db.session.commit()

        app.user1_id = user1.id
        app.user3_id = user3.id
        app.personal_token = personal_token.access_token
        app.personal_token3 = personal_token3.access_token
    return app


@pytest.fixture
def resource_fixture(app, api_app):
    """Fixture that contains the test data for models tests."""
    from flask import g, request

    from invenio_oauth2server.proxies import current_oauth2server

    # Setup API resources
    class Test0Resource(MethodView):
        def get(self):
            app.identity = g.identity
            return "success", 200

    class Test1Resource(MethodView):
        # NOTE: Method decorators are applied in reverse order
        decorators = [
            require_oauth_scopes("test:testscope"),
            require_api_auth(),
        ]

        def get(self):
            assert request.oauth.access_token
            app.identity = g.identity
            return "success", 200

        def post(self):
            assert request.oauth.access_token
            return "success", 200

    class Test2Resource(MethodView):
        @require_api_auth()
        @require_oauth_scopes("test:testscope")
        def get(self):
            assert request.oauth.access_token
            return "success", 200

        @require_api_auth()
        @require_oauth_scopes("test:testscope")
        def post(self):
            assert request.oauth.access_token
            return "success", 200

    class Test3Resource(MethodView):
        @require_api_auth()
        def post(self):
            return "success", 200

    class Test4Resource(MethodView):
        @require_api_auth(allow_anonymous=True)
        def get(self):
            from flask_login import current_user

            return str(current_user.get_id()), 200

        def post(self):
            from flask_login import current_user

            return str(current_user.get_id()), 200

    # Register API resources
    api_app.add_url_rule(
        "/test0/identitytestcase/",
        view_func=Test0Resource.as_view("test0resource"),
    )
    api_app.add_url_rule(
        "/test1/decoratorstestcase/",
        view_func=Test1Resource.as_view("test1resource"),
    )
    api_app.add_url_rule(
        "/test2/decoratorstestcase/",
        view_func=Test2Resource.as_view("test2resource"),
    )
    # This one is a UI resource using login_required
    api_app.add_url_rule(
        "/test3/loginrequiredstestcase/",
        view_func=Test3Resource.as_view("test3resource"),
    )
    api_app.add_url_rule(
        "/test4/allowanonymous/",
        view_func=Test4Resource.as_view("test4resource"),
    )

    datastore = app.extensions["security"].datastore
    with app.app_context():
        # Register a test scope
        current_oauth2server.register_scope(
            Scope("test:testscope", group="Test", help_text="Test scope")
        )
        with db.session.begin_nested():
            app.user = datastore.create_user(
                email="info@inveniosoftware.org",
                password="tester",
                active=True,
            )

        # Create tokens
        app.user_id = app.user.id
        app.token = Token.create_personal(
            "test-", app.user.id, scopes=["test:testscope"], is_internal=True
        ).access_token
        app.token_noscope = Token.create_personal(
            "test-", app.user.id, scopes=[], is_internal=True
        ).access_token
        db.session.commit()

    with api_app.test_request_context():
        app.url_for_test3resource = url_for("test3resource")

    with api_app.test_request_context():
        app.url_for_test0resource = url_for("test0resource")
        app.url_for_test1resource = url_for("test1resource")
        app.url_for_test2resource = url_for("test2resource")
        app.url_for_test4resource = url_for("test4resource")
        app.url_for_test0resource_token = url_for(
            "test0resource", access_token=app.token
        )
        app.url_for_test1resource_token = url_for(
            "test1resource", access_token=app.token
        )
        app.url_for_test2resource_token = url_for(
            "test2resource", access_token=app.token
        )
        app.url_for_test1resource_token_noscope = url_for(
            "test1resource", access_token=app.token_noscope
        )
        app.url_for_test2resource_token_noscope = url_for(
            "test2resource", access_token=app.token_noscope
        )
        app.url_for_test4resource_token = url_for(
            "test4resource", access_token=app.token
        )

    return app



# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Render semantic ui styling for forms."""

from wtforms.widgets import HTMLString


class SelectSUI(object):
    """Renders a select field with Semantic UI styling."""

    def __init__(self, multiple=False):
        """Init."""
        self.multiple = multiple

    def __call__(self, field, **kwargs):
        """Render select field."""
        html = ['<div class="ui fluid selection dropdown">']
        html.append('<input type="hidden" name="{0}">'.format(field.name))
        html.append('<i class="dropdown icon"></i>')
        items_html = []
        default_text_html = []
        for val, label, selected in field.iter_choices():
            if selected:
                default_text_html = [
                    '<div class="text">{0}</div><div class="menu">'.format(label)
                ]
            items_html.append(self.render_option(val, label, selected))
        items_html.append("</div></div>")
        html = html + default_text_html + items_html
        return HTMLString("".join(html))

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        """Render option."""
        return HTMLString(
            '<div class="item" data-value="{0}">{1}</div>'.format(value, label)
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test settings views."""

from __future__ import absolute_import, print_function

from flask import url_for
from flask_babelex import gettext as _
from helpers import login

from invenio_oauth2server.models import Client, Token


def test_personal_token_management(settings_fixture):
    """Test managing personal tokens through the views."""
    app = settings_fixture
    with app.test_request_context():
        with app.test_client() as client:
            login(client)

            # Non-existing token should return 404
            resp = client.get(
                url_for("invenio_oauth2server_settings.token_view", token_id=1000)
            )
            resp.status_code == 404

            # Get the new token form
            resp = client.get(url_for("invenio_oauth2server_settings.token_new"))
            resp.status_code == 200
            assert _("New personal access token") in str(resp.get_data())
            assert '<label for="scopes-test:scope"' in str(resp.get_data())
            assert '<label for="scopes-test:scope2"' in str(resp.get_data())

            # Create a new token with invalid form data
            resp = client.post(
                url_for("invenio_oauth2server_settings.token_new"),
                data={
                    "name": "x" * (40 + 1),  # max length is 40
                },
                follow_redirects=True,
            )
            assert resp.status_code == 200
            assert "name must be less than 40 char" in str(resp.get_data())

            # Create a new token
            resp = client.post(
                url_for("invenio_oauth2server_settings.token_new"),
                data={
                    "name": "Test_Token",
                },
                follow_redirects=True,
            )
            assert resp.status_code == 200
            assert "Personal access token / Test_Token" in str(resp.get_data())
            assert "test:scope" in str(resp.get_data())
            assert "test:scope2" in str(resp.get_data())

            token = Token.query.first()

            # Rename the token
            resp = client.post(
                url_for("invenio_oauth2server_settings.token_view", token_id=token.id),
                data=dict(name="Test_Token_Renamed"),
            )
            assert resp.status_code == 200
            assert "Test_Token_Renamed" in str(resp.get_data())

            # Token should be visible on index
            resp = client.get(url_for("invenio_oauth2server_settings.index"))
            assert resp.status_code == 200
            assert "Test_Token_Renamed" in str(resp.get_data())

            # Delete the token
            resp = client.post(
                url_for("invenio_oauth2server_settings.token_view", token_id=1),
                data=dict(delete=True),
                follow_redirects=True,
            )
            assert resp.status_code == 200
            # Token should no longer exist on index
            assert "Test_Token_Renamed" not in str(resp.get_data())


def test_authorized_app_revocation(developer_app_fixture):
    """Test managing authorized application tokens through the views."""
    app = developer_app_fixture
    with app.test_request_context():
        with app.test_client() as client:
            login(client)

            # Check that there is a single token for the authorized application
            assert Token.query.count() == 1

            # Check that the authorized application is visible on index view
            resp = client.get(url_for("invenio_oauth2server_settings.index"))
            assert resp.status_code == 200
            assert "Test description" in str(resp.get_data())
            assert "Test name" in str(resp.get_data())

            # Revoke the authorized application token
            resp = client.get(
                url_for("invenio_oauth2server_settings.token_revoke", token_id=1),
                follow_redirects=True,
            )
            assert resp.status_code == 200
            # Authorized application should no longer exist on index
            assert "Test description" not in str(resp.get_data())
            assert "Test name" not in str(resp.get_data())
            # Check that the authorized application token was actually deleted
            assert Token.query.count() == 0


def test_client_management(settings_fixture):
    """Test managing clients through the views."""
    app = settings_fixture
    with app.test_request_context():
        with app.test_client() as client:
            login(client)

            # Non-existing client should return 404
            resp = client.get(
                url_for("invenio_oauth2server_settings.client_view", client_id=1000)
            )
            assert resp.status_code == 404

            # Create a new client
            resp = client.post(
                url_for("invenio_oauth2server_settings.client_new"),
                data=dict(
                    name="Test_Client",
                    description="Test description for Test_Client.",
                    website="http://inveniosoftware.org/",
                    redirect_uris=url_for(
                        "invenio_oauth2server_settings.index", _external=True
                    ),
                    is_confditential=1,
                ),
                follow_redirects=True,
            )
            assert resp.status_code == 200
            assert "Application / Test_Client" in str(resp.get_data())
            test_client = Client.query.first()
            assert test_client.client_id in str(resp.get_data())

            # Client should be visible on index
            resp = client.get(url_for("invenio_oauth2server_settings.index"))
            assert resp.status_code == 200
            assert "Test_Client" in str(resp.get_data())

            # Reset client secret
            original_client_secret = test_client.client_secret
            resp = client.post(
                url_for(
                    "invenio_oauth2server_settings.client_reset",
                    client_id=test_client.client_id,
                ),
                data=dict(reset="yes"),
                follow_redirects=True,
            )
            assert resp.status_code == 200
            assert test_client.client_secret in str(resp.get_data())
            assert original_client_secret not in str(resp.get_data())

            # Invalid redirect uri should error
            original_redirect_uris = test_client.redirect_uris
            resp = client.post(
                url_for(
                    "invenio_oauth2server_settings.client_view",
                    client_id=test_client.client_id,
                ),
                data=dict(
                    name="Test_Client",
                    description="Test description for Test_Client",
                    website="http://inveniosoftware.org/",
                    redirect_uris="https:/invalid",
                ),
            )
            assert resp.status_code == 200
            assert test_client.redirect_uris == original_redirect_uris

            # Modify the client
            resp = client.post(
                url_for(
                    "invenio_oauth2server_settings.client_view",
                    client_id=test_client.client_id,
                ),
                data=dict(
                    name="Modified_Name",
                    description="Modified Description",
                    website="http://modified-url.org",
                    redirect_uris="https://example.org",
                ),
            )
            assert resp.status_code == 200
            assert "Modified_Name" in str(resp.get_data())
            assert "Modified Description" in str(resp.get_data())
            assert "http://modified-url.org" in str(resp.get_data())

            # Delete the client
            resp = client.post(
                url_for(
                    "invenio_oauth2server_settings.client_view",
                    client_id=test_client.client_id,
                ),
                follow_redirects=True,
                data=dict(delete=True),
            )
            assert resp.status_code == 200
            assert test_client.name not in str(resp.get_data())



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""OAuth Server Settings Blueprint."""

from __future__ import absolute_import

from functools import wraps

from flask import Blueprint, abort, redirect, render_template, request, session, url_for
from flask_babelex import lazy_gettext as _
from flask_breadcrumbs import register_breadcrumb
from flask_login import current_user, login_required
from flask_menu import register_menu
from invenio_db import db
from invenio_theme.proxies import current_theme_icons
from speaklater import make_lazy_string

from ..forms import ClientForm, TokenForm
from ..models import Client, Token
from ..proxies import current_oauth2server

blueprint = Blueprint(
    "invenio_oauth2server_settings",
    __name__,
    url_prefix="/account/settings/applications",
    static_folder="../static",
    template_folder="../templates",
)


#
# Decorator
#
def client_getter():
    """Decorator to retrieve Client object and check user permission."""

    def wrapper(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            if "client_id" not in kwargs:
                abort(500)

            client = Client.query.filter_by(
                client_id=kwargs.pop("client_id"),
                user_id=current_user.get_id(),
            ).first()

            if client is None:
                abort(404)

            return f(client, *args, **kwargs)

        return decorated

    return wrapper


def token_getter(is_personal=True, is_internal=False):
    """Decorator to retrieve Token object and check user permission.

    :param is_personal: Search for a personal token. (Default: ``True``)
    :param is_internal: Search for a internal token. (Default: ``False``)
    """

    def wrapper(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            if "token_id" not in kwargs:
                abort(500)

            token = Token.query.filter_by(
                id=kwargs.pop("token_id"),
                user_id=current_user.get_id(),
                is_personal=is_personal,
                is_internal=is_internal,
            ).first()

            if token is None:
                abort(404)

            return f(token, *args, **kwargs)

        return decorated

    return wrapper


#
# Views
#
@blueprint.route("/", methods=["GET", "POST"])
@login_required
@register_menu(
    blueprint,
    "settings.applications",
    _(
        "%(icon)s Applications",
        icon=make_lazy_string(lambda: f'<i class="{current_theme_icons.codepen}"></i>'),
    ),
    order=5,
    active_when=lambda: request.endpoint.startswith("invenio_oauth2server_settings."),
)
@register_breadcrumb(blueprint, "breadcrumbs.settings.applications", _("Applications"))
def index():
    """List user tokens."""
    clients = Client.query.filter_by(
        user_id=current_user.get_id(),
        is_internal=False,
    ).all()

    tokens = (
        Token.query.options(db.joinedload("client"))
        .filter(
            Token.user_id == current_user.get_id(),
            Token.is_personal == True,  # noqa
            Token.is_internal == False,
            Client.is_internal == True,
        )
        .all()
    )

    authorized_apps = (
        Token.query.options(db.joinedload("client"))
        .filter(
            Token.user_id == current_user.get_id(),
            Token.is_personal == False,  # noqa
            Token.is_internal == False,
            Client.is_internal == False,
        )
        .all()
    )

    return render_template(
        "invenio_oauth2server/settings/index.html",
        clients=clients,
        tokens=tokens,
        authorized_apps=authorized_apps,
    )


@blueprint.route("/clients/new/", methods=["GET", "POST"])
@login_required
@register_breadcrumb(
    blueprint, "breadcrumbs.settings.applications.client_new", _("New")
)
def client_new():
    """Create new client."""
    form = ClientForm(request.form)

    if form.validate_on_submit():
        c = Client(user_id=current_user.get_id())
        c.gen_salt()
        form.populate_obj(c)
        db.session.add(c)
        db.session.commit()
        return redirect(url_for(".client_view", client_id=c.client_id))

    return render_template(
        "invenio_oauth2server/settings/client_new.html",
        form=form,
    )


@blueprint.route("/clients/<string:client_id>/", methods=["GET", "POST"])
@login_required
@register_breadcrumb(
    blueprint, "breadcrumbs.settings.applications.client_edit", _("Edit")
)
@client_getter()
def client_view(client):
    """Show client's detail."""
    if request.method == "POST" and "delete" in request.form:
        db.session.delete(client)
        db.session.commit()
        return redirect(url_for(".index"))

    form = ClientForm(request.form, obj=client)
    if form.validate_on_submit():
        form.populate_obj(client)
        db.session.commit()

    return render_template(
        "invenio_oauth2server/settings/client_view.html",
        client=client,
        form=form,
    )


@blueprint.route("/clients/<string:client_id>/reset/", methods=["POST"])
@login_required
@client_getter()
def client_reset(client):
    """Reset client's secret."""
    if request.form.get("reset") == "yes":
        client.reset_client_secret()
        db.session.commit()
    return redirect(url_for(".client_view", client_id=client.client_id))


#
# Token views
#
@blueprint.route("/tokens/new/", methods=["GET", "POST"])
@login_required
@register_breadcrumb(blueprint, "breadcrumbs.settings.applications.token_new", _("New"))
def token_new():
    """Create new token."""
    form = TokenForm(request.form)
    form.scopes.choices = current_oauth2server.scope_choices()

    if form.validate_on_submit():
        t = Token.create_personal(
            form.data["name"], current_user.get_id(), scopes=form.scopes.data
        )
        db.session.commit()
        session["show_personal_access_token"] = True
        return redirect(url_for(".token_view", token_id=t.id))

    if len(current_oauth2server.scope_choices()) == 0:
        del form.scopes

    return render_template(
        "invenio_oauth2server/settings/token_new.html",
        form=form,
    )


@blueprint.route("/tokens/<string:token_id>/", methods=["GET", "POST"])
@login_required
@register_breadcrumb(
    blueprint, "breadcrumbs.settings.applications.token_edit", _("Edit")
)
@token_getter()
def token_view(token):
    """Show token details."""
    if request.method == "POST" and "delete" in request.form:
        db.session.delete(token)
        db.session.commit()
        return redirect(url_for(".index"))

    show_token = session.pop("show_personal_access_token", False)

    form = TokenForm(request.form, name=token.client.name, scopes=token.scopes)
    form.scopes.choices = current_oauth2server.scope_choices()
    form.scopes.data = token.scopes

    if form.validate_on_submit():
        token.client.name = form.data["name"]
        token.scopes = form.data["scopes"]
        db.session.commit()

    if len(current_oauth2server.scope_choices()) == 0:
        del form.scopes

    return render_template(
        "invenio_oauth2server/settings/token_view.html",
        token=token,
        form=form,
        show_token=show_token,
    )


@blueprint.route(
    "/tokens/<string:token_id>/revoke/",
    methods=[
        "GET",
    ],
)
@login_required
@token_getter(is_personal=False, is_internal=False)
def token_revoke(token):
    """Revoke Authorized Application token."""
    db.session.delete(token)
    db.session.commit()
    return redirect(url_for(".index"))


@blueprint.route(
    "/tokens/<string:token_id>/view/",
    methods=[
        "GET",
    ],
)
@login_required
@token_getter(is_personal=False, is_internal=False)
def token_permission_view(token):
    """Show permission garanted to authorized application token."""
    scopes = [current_oauth2server.scopes[x] for x in token.scopes]
    return render_template(
        "invenio_oauth2server/settings/token_permission_view.html",
        token=token,
        scopes=scopes,
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test OAuth2Server providers."""

from __future__ import absolute_import, print_function

from datetime import datetime, timedelta

import pytest
from flask import json, url_for
from helpers import login, parse_redirect
from invenio_accounts.models import User
from invenio_db import db

from invenio_oauth2server.ext import InvenioOAuth2ServerREST
from invenio_oauth2server.models import Client, Token


def test_client_salt(provider_fixture):
    app = provider_fixture
    with app.app_context():
        with db.session.begin_nested():
            client = Client(
                name="Test something",
                is_confidential=True,
                user_id=1,
            )

            client.gen_salt()
            assert (
                len(client.client_id) == app.config["OAUTH2SERVER_CLIENT_ID_SALT_LEN"]
            )
            assert (
                len(client.client_secret)
                == app.config["OAUTH2SERVER_CLIENT_SECRET_SALT_LEN"]
            )

            db.session.add(client)

        with db.session.begin_nested():
            db.session.delete(client)


def test_invalid_authorize_requests(provider_fixture):
    app = provider_fixture
    # First login on provider site
    with app.test_request_context():
        redirect_uri = url_for("oauth2test.authorized", _external=True)

        with app.test_client() as client:
            login(client)

            for client_id in ["dev", "confidential"]:
                scope = "test:scope"
                response_type = "code"

                error_url = url_for("invenio_oauth2server.errors")

                # Valid request authorize request
                r = client.get(
                    url_for("invenio_oauth2server.authorize"),
                    data={
                        "redirect_uri": redirect_uri,
                        "scope": scope,
                        "response_type": response_type,
                        "client_id": client_id,
                    },
                )
                assert r.status_code == 200

                # Invalid scope
                r = client.get(
                    url_for(
                        "invenio_oauth2server.authorize",
                        redirect_uri=redirect_uri,
                        scope="INVALID",
                        response_type=response_type,
                        client_id=client_id,
                    )
                )
                assert r.status_code == 302
                next_url, data = parse_redirect(r.location)
                assert data["error"] == "invalid_scope"
                assert next_url == redirect_uri

                # Invalid response type
                r = client.get(
                    url_for(
                        "invenio_oauth2server.authorize",
                        redirect_uri=redirect_uri,
                        scope=scope,
                        response_type="invalid",
                        client_id=client_id,
                    )
                )
                assert r.status_code == 302
                next_url, data = parse_redirect(r.location)
                assert data["error"] == "unsupported_response_type"
                assert next_url == redirect_uri

                # Missing response_type
                r = client.get(
                    url_for(
                        "invenio_oauth2server.authorize",
                        redirect_uri=redirect_uri,
                        scope=scope,
                        client_id=client_id,
                    )
                )
                assert r.status_code == 302
                next_url, data = parse_redirect(r.location)
                assert data["error"] == "invalid_request"
                assert next_url == redirect_uri

                # Duplicate parameter
                r = client.get(
                    url_for(
                        "invenio_oauth2server.authorize",
                        redirect_uri=redirect_uri,
                        scope=scope,
                        response_type="invalid",
                        client_id=client_id,
                    )
                    + "&client_id={0}".format(client_id)
                )
                assert r.status_code == 302
                next_url, data = parse_redirect(r.location)
                assert data["error"] == "invalid_request"
                assert next_url == error_url

                # Invalid client_id
                r = client.get(
                    url_for(
                        "invenio_oauth2server.authorize",
                        redirect_uri=redirect_uri,
                        scope=scope,
                        response_type=response_type,
                        client_id="invalid",
                    )
                )
                assert r.status_code == 302
                next_url, data = parse_redirect(r.location)
                assert data["error"] == "invalid_request"
                assert error_url in next_url

                r = client.get(next_url, query_string=data)
                assert "invalid_request" in str(r.data)

                # Invalid redirect uri
                r = client.get(
                    url_for(
                        "invenio_oauth2server.authorize",
                        redirect_uri="http://localhost/",
                        scope=scope,
                        response_type=response_type,
                        client_id=client_id,
                    )
                )
                assert r.status_code == 302
                next_url, data = parse_redirect(r.location)
                assert data["error"] == "invalid_request"
                assert error_url in next_url


def test_refresh_flow(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        base_url = "https://{0}".format(app.config["SERVER_NAME"])
        redirect_uri = url_for("oauth2test.authorized", _external=True)

        with app.test_client() as client:
            # First login on provider site
            login(client)

            data = dict(
                redirect_uri=redirect_uri,
                scope="test:scope",
                response_type="code",
                client_id="confidential",
                state="mystate",
            )

            r = client.get(url_for("invenio_oauth2server.authorize", **data))
            assert r.status_code == 200

            data["confirm"] = "yes"
            data["scope"] = "test:scope"
            data["state"] = "mystate"

            # Obtain one time code
            r = client.post(url_for("invenio_oauth2server.authorize"), data=data)
            r.status_code == 302
            next_url, res_data = parse_redirect(r.location)
            assert res_data["code"]
            assert res_data["state"] == "mystate"

            # Exchange one time code for access token
            r = client.post(
                url_for("invenio_oauth2server.access_token"),
                data=dict(
                    client_id="confidential",
                    client_secret="confidential",
                    grant_type="authorization_code",
                    code=res_data["code"],
                ),
            )
            assert r.status_code == 200
            json_resp = json.loads(r.get_data())
            assert json_resp["access_token"]
            assert json_resp["refresh_token"]
            assert json_resp["scope"] == "test:scope"
            assert json_resp["token_type"] == "Bearer"
            assert json_resp["user"]["id"]
            refresh_token = json_resp["refresh_token"]
            old_access_token = json_resp["access_token"]

            # Access token valid
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=old_access_token)
            )
            assert r.status_code == 200

            # Obtain new access token with refresh token
            r = client.post(
                url_for("invenio_oauth2server.access_token"),
                data=dict(
                    client_id="confidential",
                    client_secret="confidential",
                    grant_type="refresh_token",
                    refresh_token=refresh_token,
                ),
            )
            r.status_code == 200
            json_resp = json.loads(r.get_data())
            assert json_resp["access_token"]
            assert json_resp["refresh_token"]
            assert json_resp["access_token"] != old_access_token
            assert json_resp["refresh_token"] != refresh_token
            assert json_resp["scope"] == "test:scope"
            assert json_resp["token_type"] == "Bearer"
            assert json_resp["user"]["id"]

            # New access token valid
            r = client.get(
                url_for(
                    "invenio_oauth2server.info", access_token=json_resp["access_token"]
                )
            )
            assert r.status_code == 200

            # Old access token no longer valid
            r = client.get(
                url_for(
                    "invenio_oauth2server.info",
                    access_token=old_access_token,
                ),
                base_url=base_url,
            )
            assert r.status_code == 401


def web_auth_flow(provider_fixture):
    app = provider_fixture
    # Go to login - should redirect to OAuth 2 server for login an
    # authorization
    with app.app_context():
        with app.test_client() as client:
            r = client.get("/oauth2test/test-ping")

            # First login on provider site
            login(client)

            r = client.get("/oauth2test/login")
            assert r.status_code == 302
            next_url, data = parse_redirect(r.location)

            # Authorize page
            r = client.get(next_url, query_string=data)
            assert r.status_code == 200

            # User confirms request
            data["confirm"] = "yes"
            data["scope"] = "test:scope"
            data["state"] = ""
            r = client.post(next_url, data=data)
            assert r.status_code == 302
            next_url, data = parse_redirect(r.location)
            assert next_url == "http://{0}/oauth2test/authorized".format(
                app.config["SERVER_NAME"]
            )
            assert "code" in data

            # User is redirected back to client site.
            # - The client view /oauth2test/authorized will in the
            #   background fetch the access token.
            r = client.get(next_url, query_string=data)
            assert r.status_code == 200

            # Authentication flow has now been completed, and the access
            # token can be used to access protected resources.
            r = client.get("/oauth2test/test-ping")
            assert r.status_code == 200
            assert json.loads(r.get_data()) == dict(ping="pong")

            # Authentication flow has now been completed, and the access
            # token can be used to access protected resources.
            r = client.get("/oauth2test/test-ping")
            assert r.status_code == 200
            assert json.loads(r.get_data()) == dict(ping="pong")

            r = client.get("/oauth2test/test-info")
            assert r.status_code == 200
            json_resp = json.loads(r.get_data())
            assert json_resp["client"] == "confidential"
            assert json_resp["user"] == app.user1_id
            assert json_resp["scopes"] == ["test:scope"]

            # Access token doesn't provide access to this URL.
            r = client.get(
                "/oauth2test/test-invalid",
                base_url="http://{0}".format(app.config["SERVER_NAME"]),
            )
            assert r.status_code == 401

            # # Now logout
            r = client.get("/oauth2test/logout")
            assert r.status_code == 200
            assert r.data == "logout"

            # And try to access the information again
            r = client.get("/oauth2test/test-ping")
            assert r.status_code == 403


def test_implicit_flow(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        redirect_uri = url_for("oauth2test.authorized", _external=True)

        with app.test_client() as client:
            # First login on provider site
            login(client)

            for client_id in ["dev", "confidential"]:
                data = dict(
                    redirect_uri=redirect_uri,
                    response_type="token",  # For implicit grant type
                    client_id=client_id,
                    scope="test:scope",
                    state="teststate",
                )

                # Authorize page
                r = client.get(
                    url_for("invenio_oauth2server.authorize", **data),
                    follow_redirects=True,
                )
                assert r.status_code == 200

                # User confirms request
                data["confirm"] = "yes"
                data["scope"] = "test:scope"
                data["state"] = "teststate"

                r = client.post(url_for("invenio_oauth2server.authorize"), data=data)
                assert r.status_code == 302
                # Important - access token exists in URI fragment and must not
                # be sent to the client.
                next_url, data = parse_redirect(r.location, parse_fragment=True)

                assert data["access_token"]
                assert data["token_type"] == "Bearer"
                assert data["state"] == "teststate"
                assert data["scope"] == "test:scope"
                assert data.get("refresh_token") is None
                assert next_url == redirect_uri

                # Authentication flow has now been completed, and the client
                # can use the access token to make request to the provider.
                r = client.get(
                    url_for(
                        "invenio_oauth2server.info", access_token=data["access_token"]
                    )
                )
                assert r.status_code == 200
                assert json.loads(r.get_data()).get("client") == client_id
                assert json.loads(r.get_data()).get("user") == app.user1_id
                assert json.loads(r.get_data()).get("scopes") == ["test:scope"]


def test_client_flow(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        with app.test_client() as client:
            data = dict(
                client_id="dev",
                client_secret="dev",  # A public client should NOT do this!
                grant_type="client_credentials",
                scope="test:scope",
            )

            # Public clients are not allowed to use
            # grant_type=client_credentials
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 401
            assert json.loads(r.get_data()).get("error") == "invalid_client"

            data = dict(
                client_id="confidential",
                client_secret="confidential",
                grant_type="client_credentials",
                scope="test:scope",
            )

            # Retrieve access token using client_credentials
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 200

            data = json.loads(r.get_data())
            assert data["access_token"]
            assert data["token_type"] == "Bearer"
            assert data["scope"] == "test:scope"
            assert data["user"]["id"]
            assert data.get("refresh_token") is None

            # Authentication flow has now been completed, and the client can
            # use the access token to make request to the provider.
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=data["access_token"])
            )
            assert r.status_code == 200
            assert json.loads(r.get_data()).get("client") == "confidential"
            assert json.loads(r.get_data()).get("user") == app.user1_id
            assert json.loads(r.get_data()).get("scopes") == ["test:scope"]

            data = dict(
                client_id="confidential-user-inactive",
                client_secret="confidential-user-inactive",
                grant_type="client_credentials",
                scope="test:scope",
            )

            # Retrieve access token using client_credentials
            # belonging to an inactive user
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 401

            # Activate the client owner temporarily and request the token
            user = User.query.filter_by(email="inactive@inveniosoftware.org").first()
            user.active = True
            db.session.commit()

            # Retrieve access token using client_credentials
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 200

            data = json.loads(r.get_data())
            assert data["access_token"]
            assert data["token_type"] == "Bearer"
            assert data["scope"] == "test:scope"
            assert data["user"]["id"]
            assert data.get("refresh_token") is None

            # As before, authentication flow has now been completed, client can
            # use the access token to make request to the provider.
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=data["access_token"])
            )
            assert r.status_code == 200
            assert (
                json.loads(r.get_data()).get("client") == "confidential-user-inactive"
            )
            assert json.loads(r.get_data()).get("user") == app.user3_id
            assert json.loads(r.get_data()).get("scopes") == ["test:scope"]

            # Inactivate the client owner again
            user = User.query.filter_by(email="inactive@inveniosoftware.org").first()
            user.active = False
            db.session.commit()

            # Token is invalid because the client owner account has been
            # deactivated
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=data["access_token"])
            )
            assert r.status_code == 401


def test_auth_flow_denied(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        redirect_uri = url_for("oauth2test.authorized", _external=True)
        with app.test_client() as client:
            # First login on provider site
            login(client)

            r = client.get(url_for("oauth2test.login"))
            assert r.status_code == 302
            next_url, data = parse_redirect(r.location)

            # Authorize page
            r = client.get(next_url, query_string=data)
            assert r.status_code == 200

            # User rejects request
            data["confirm"] = "no"
            data["scope"] = "test:scope"
            data["state"] = ""

            r = client.post(next_url, data=data)
            assert r.status_code == 302
            next_url, data = parse_redirect(r.location)
            assert next_url == redirect_uri
            assert data.get("error") == "access_denied"

        # Returned
        r = client.get(next_url, query_string=data)
        assert r.status_code == 200
        assert r.data == b"Access denied: error=access_denied"


def test_personal_access_token(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        with app.test_client() as client:
            r = client.get(
                url_for("invenio_oauth2server.ping"),
                query_string="access_token={0}".format(app.personal_token),
            )
            assert r.status_code == 200
            assert json.loads(r.get_data()) == dict(ping="pong")

            # Access token is not valid for this scope
            r = client.get(
                url_for("invenio_oauth2server.info"),
                query_string="access_token={0}".format(app.personal_token),
            )
            assert r.status_code == 401

            # Access token of a user that is inactive
            r = client.get(
                url_for("invenio_oauth2server.ping"),
                query_string="access_token={0}".format(app.personal_token3),
            )
            assert r.status_code == 401


def test_resource_auth_methods(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        with app.test_client() as client:
            # Query string
            r = client.get(
                url_for("invenio_oauth2server.ping"),
                query_string="access_token={0}".format(app.personal_token),
            )
            r.status_code == 200
            assert json.loads(r.get_data()) == dict(ping="pong")

            # POST request body
            r = client.post(
                url_for("invenio_oauth2server.ping"),
                data=dict(access_token=app.personal_token),
            )
            assert r.status_code == 200
            assert json.loads(r.get_data()) == dict(ping="pong")

            # Authorization Header
            r = client.get(
                url_for("invenio_oauth2server.ping"),
                headers=[("Authorization", "Bearer {0}".format(app.personal_token))],
            )
            assert r.status_code == 200
            assert json.loads(r.get_data()) == dict(ping="pong")


@pytest.mark.parametrize(
    ("query_string", "valid_qs"),
    (
        # Valid query strings
        ("q=RegularArg", True),
        ("a=1&b=2&c=3", True),
        ("q=text+with+spaces", True),
        ("q=title:TheTitle", True),
        ("q=properly%20encoded%24", True),
        # Invalid query strings
        ("$type=search", False),
        ("q=Joan+D'Arc", False),
        ("with regular spaces", False),
        ("json_data={a: 42}", False),
        ("array=[1, 2, 3]", False),
    ),
)
def test_oauthlib_urldecoding_issue(api_app_with_test_view, query_string, valid_qs):
    app = api_app_with_test_view
    with app.test_request_context():
        with app.test_client() as client:
            # Remove '/api' since our client is not aware of the the WSGI mount
            test_url = url_for("test").replace("/api", "")
            if valid_qs:
                assert (
                    client.get(test_url, query_string=query_string).status_code == 200
                )
            else:
                assert (
                    client.get(test_url, query_string=query_string).status_code == 400
                )


def test_oauthlib_monkeypatch(api_app_with_test_view):
    app = api_app_with_test_view
    InvenioOAuth2ServerREST.monkeypatch_oauthlib_urlencode_chars("$:")

    with app.test_request_context():
        with app.test_client() as client:
            # Remove '/api' since our client is not aware of the the WSGI mount
            test_url = url_for("test").replace("/api", "")

            # '$' and ':' should be valid characters after this patch
            assert client.get(test_url, query_string="$type:search").status_code == 200

            # '=' is not considered a valid character after this patch
            assert client.get(test_url, query_string="q=RegularArg").status_code == 400


def test_settings_index(provider_fixture):
    app = provider_fixture
    with app.test_request_context():
        with app.test_client() as client:
            # Create a remote account (linked account)
            client.get(
                url_for("invenio_oauth2server_settings.index"),
                follow_redirects=True,
            )
            # assert 401 == r.status_code
            login(client)

            res = client.get(
                url_for("invenio_oauth2server_settings.index"),
            )
            assert 200 == res.status_code

            res = client.get(
                url_for("invenio_oauth2server_settings.client_new"),
            )
            assert 200 == res.status_code

            # Valid POST
            res = client.post(
                url_for("invenio_oauth2server_settings.client_new"),
                data=dict(
                    name="Test",
                    description="Test description",
                    website="http://inveniosoftware.org",
                    is_confidential=True,
                    redirect_uris="http://localhost/oauth/authorized/",
                ),
            )
            assert 302 == res.status_code

            # Invalid redirect_uri (must be https)
            res = client.post(
                url_for("invenio_oauth2server_settings.client_new"),
                data=dict(
                    name="Test",
                    description="Test description",
                    website="http://inveniosoftware.org",
                    is_confidential=True,
                    redirect_uris="http://example.org/oauth/authorized/",
                ),
            )
            assert 200 == res.status_code

            # Valid
            res = client.post(
                url_for("invenio_oauth2server_settings.client_new"),
                data=dict(
                    name="Test",
                    description="Test description",
                    website="http://inveniosoftware.org",
                    is_confidential=True,
                    redirect_uris="https://example.org/oauth/authorized/\n"
                    "http://localhost:4000/oauth/authorized/",
                ),
            )
            assert 302 == res.status_code


def test_info_not_accessible_in_production(provider_fixture):
    """Info route should not be available in production mode."""
    app = provider_fixture
    with app.test_request_context():
        with app.test_client() as client:
            app.config.update(
                DEBUG=False,
                TESTING=False,
            )
            data = dict(
                client_id="dev",
                client_secret="dev",  # A public client should NOT do this!
                grant_type="client_credentials",
                scope="test:scope",
            )

            # Public clients are not allowed to use
            # grant_type=client_credentials
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 401
            assert json.loads(r.get_data()).get("error") == "invalid_client"

            data = dict(
                client_id="confidential",
                client_secret="confidential",
                grant_type="client_credentials",
                scope="test:scope",
            )

            # Retrieve access token using client_credentials
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 200

            data = json.loads(r.get_data())
            assert data["access_token"]
            assert data["token_type"] == "Bearer"
            assert data["scope"] == "test:scope"
            assert data.get("refresh_token") is None
            assert data["user"]["id"]

            # Authentication flow has now been completed, and the client can
            # use the access token to make request to the provider.
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=data["access_token"])
            )
            assert r.status_code == 404


def test_expired_refresh_flow(provider_fixture):
    """Test refresh flow with an expired token."""
    app = provider_fixture
    # First login on provider site
    with app.test_request_context():
        with app.test_client() as client:
            login(client)

            data = dict(
                redirect_uri=url_for("oauth2test.authorized", _external=True),
                scope="test:scope",
                response_type="code",
                client_id="confidential",
                state="mystate",
            )

            r = client.get(url_for("invenio_oauth2server.authorize", **data))
            assert r.status_code == 200

            data["confirm"] = "yes"
            data["scope"] = "test:scope"
            data["state"] = "mystate"

            # Obtain one time code
            r = client.post(url_for("invenio_oauth2server.authorize"), data=data)
            assert r.status_code == 302
            next_url, res_data = parse_redirect(r.location)
            assert res_data["code"]
            assert res_data["state"] == "mystate"

            # Exchange one time code for access token
            r = client.post(
                url_for("invenio_oauth2server.access_token"),
                data=dict(
                    client_id="confidential",
                    client_secret="confidential",
                    grant_type="authorization_code",
                    code=res_data["code"],
                ),
            )
            assert r.status_code == 200
            assert json.loads(r.get_data())["access_token"]
            assert json.loads(r.get_data())["refresh_token"]
            assert json.loads(r.get_data())["expires_in"] > 0
            assert json.loads(r.get_data())["scope"] == "test:scope"
            assert json.loads(r.get_data())["token_type"] == "Bearer"
            assert json.loads(r.get_data())["user"]["id"]
            refresh_token = json.loads(r.get_data())["refresh_token"]
            old_access_token = json.loads(r.get_data())["access_token"]

            # Access token valid
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=old_access_token)
            )
            assert r.status_code == 200

            Token.query.filter_by(access_token=old_access_token).update(
                dict(expires=datetime.utcnow() - timedelta(seconds=1))
            )
            db.session.commit()

            # Access token is expired
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=old_access_token)
            )
            assert r.status_code == 401

            # Obtain new access token with refresh token
            r = client.post(
                url_for("invenio_oauth2server.access_token"),
                data=dict(
                    client_id="confidential",
                    client_secret="confidential",
                    grant_type="refresh_token",
                    refresh_token=refresh_token,
                ),
            )
            assert r.status_code == 200
            resp_json = json.loads(r.get_data())
            assert resp_json["access_token"]
            assert resp_json["refresh_token"]
            assert resp_json["expires_in"] > 0
            assert resp_json["access_token"] != old_access_token
            assert resp_json["refresh_token"] != refresh_token
            assert resp_json["scope"] == "test:scope"
            assert resp_json["token_type"] == "Bearer"
            assert resp_json["user"]["id"]

            # New access token valid
            r = client.get(
                url_for(
                    "invenio_oauth2server.info", access_token=resp_json["access_token"]
                )
            )
            assert r.status_code == 200

            # Old access token no longer valid
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=old_access_token)
            )
            assert r.status_code == 401


def test_not_allowed_public_refresh_flow(provider_fixture):
    """Public token should not allow refreshing."""
    app = provider_fixture
    # First login on provider site
    with app.test_request_context():
        with app.test_client() as client:
            login(client)

            data = dict(
                redirect_uri=url_for("oauth2test.authorized", _external=True),
                scope="test:scope",
                response_type="code",
                client_id="dev",
                state="mystate",
            )

            r = client.get(url_for("invenio_oauth2server.authorize", **data))
            assert r.status_code == 200

            data["confirm"] = "yes"
            data["scope"] = "test:scope"
            data["state"] = "mystate"

            # Obtain one time code
            r = client.post(url_for("invenio_oauth2server.authorize"), data=data)
            assert r.status_code == 302
            next_url, res_data = parse_redirect(r.location)
            assert res_data["code"]
            assert res_data["state"] == "mystate"

            # Exchange one time code for access token
            r = client.post(
                url_for("invenio_oauth2server.access_token"),
                data=dict(
                    client_id="dev",
                    client_secret="dev",
                    grant_type="authorization_code",
                    code=res_data["code"],
                ),
            )
            assert r.status_code == 200
            json_resp = json.loads(r.get_data())
            assert json_resp["access_token"]
            assert json_resp["refresh_token"]
            assert json_resp["expires_in"] > 0
            assert json_resp["scope"] == "test:scope"
            assert json_resp["token_type"] == "Bearer"
            assert json_resp["user"]["id"]
            refresh_token = json_resp["refresh_token"]
            old_access_token = json_resp["access_token"]

            # Access token valid
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=old_access_token)
            )
            assert r.status_code == 200

            Token.query.filter_by(access_token=old_access_token).update(
                dict(expires=datetime.utcnow() - timedelta(seconds=1))
            )
            db.session.commit()

            # Access token is expired
            r = client.get(
                url_for("invenio_oauth2server.info", access_token=old_access_token),
                follow_redirects=True,
            )
            assert r.status_code == 401

            # Obtain new access token with refresh token
            r = client.post(
                url_for("invenio_oauth2server.access_token"),
                data=dict(
                    client_id="dev",
                    client_secret="dev",
                    grant_type="refresh_token",
                    refresh_token=refresh_token,
                ),
                follow_redirects=True,
            )
            # Only confidential clients can refresh expired token.
            assert r.status_code == 401


def test_password_grant_type(provider_fixture):
    app = provider_fixture
    app.config["OAUTH2SERVER_ALLOWED_GRANT_TYPES"] = ("password",)
    with app.test_request_context():
        with app.test_client() as client:
            data = dict(
                client_id="dev",
                client_secret="dev",
                grant_type="password",
                username="info@inveniosoftware.org",
                password="tester",
                scope="test:scope",
            )

            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 200
            res_data = json.loads(r.get_data())
            assert "Bearer" == res_data["token_type"]
            assert res_data["user"]["id"]

            data["password"] = "invalid"

            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 401

            data["username"] = "inactive@inveniosoftware.org"
            data["password"] = "tester3"
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 401


def test_email_scope(provider_fixture):
    """Test email scope."""
    app = provider_fixture
    with app.test_request_context():
        with app.test_client() as client:
            data = dict(
                client_id="confidential-email",
                client_secret="confidential-email",
                grant_type="client_credentials",
                scope="user:email",
            )

            # Retrieve access token using client_credentials
            r = client.post(url_for("invenio_oauth2server.access_token"), data=data)
            assert r.status_code == 200

            data = json.loads(r.get_data())
            assert data["access_token"]
            assert data["user"]["id"]
            assert data["scope"] == "user:email"
            assert data["user"]["email"] == "info@inveniosoftware.org"
            assert data["user"]["email_verified"] is False



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

from flask import Flask, url_for
from flask_admin import Admin
from invenio_db import db

from invenio_oauth2server import InvenioOAuth2Server
from invenio_oauth2server.admin import (
    ClientView,
    TokenView,
    oauth2server_clients_adminview,
    oauth2server_tokens_adminview,
)


def test_admin(models_fixture):
    """Test flask-admin interface."""
    app = models_fixture
    InvenioOAuth2Server(app)

    assert isinstance(oauth2server_tokens_adminview, dict)
    assert isinstance(oauth2server_clients_adminview, dict)

    assert "view_class" in oauth2server_tokens_adminview
    assert "view_class" in oauth2server_clients_adminview

    admin = Admin(app, name="Test")

    clients_view = oauth2server_clients_adminview.pop("view_class")
    clients_model, clients_session = oauth2server_clients_adminview.pop("args")
    clients_kwargs = oauth2server_clients_adminview.pop("kwargs")
    tokens_view = oauth2server_tokens_adminview.pop("view_class")
    tokens_model, tokens_session = oauth2server_tokens_adminview.pop("args")
    tokens_kwargs = oauth2server_tokens_adminview.pop("kwargs")
    admin.add_view(clients_view(clients_model, db.session, **clients_kwargs))
    admin.add_view(tokens_view(tokens_model, db.session, **tokens_kwargs))

    menu_items = {str(item.name): item for item in admin.menu()}
    assert "User Management" in menu_items
    assert menu_items["User Management"].is_category()

    submenu_items = {
        str(item.name): item for item in menu_items["User Management"].get_children()
    }
    assert "OAuth Applications" in submenu_items
    assert "OAuth Application Tokens" in submenu_items

    with app.test_request_context():
        token_request_url = url_for("token.index_view")
        client_request_url = url_for("client.index_view")
        client_view_url = url_for(
            "invenio_oauth2server_settings.client_view", client_id="client_test_u1c1"
        )
        client_reset_url = url_for(
            "invenio_oauth2server_settings.client_reset", client_id="client_test_u1c1"
        )
        token_view_url = url_for(
            "invenio_oauth2server_settings.token_view", token_id="1"
        )
        token_revoke_url = url_for(
            "invenio_oauth2server_settings.token_revoke", token_id="1"
        )

    with app.app_context():
        with app.test_client() as client:
            res = client.get(token_request_url, follow_redirects=True)
            assert res.status_code == 200
            res = client.get(client_request_url, follow_redirects=True)
            assert res.status_code == 200
            res = client.get(client_view_url, follow_redirects=True)
            assert res.status_code == 200
            res = client.post(client_reset_url, follow_redirects=True)
            assert res.status_code == 200
            res = client.get(token_view_url, follow_redirects=True)
            assert res.status_code == 200
            res = client.post(token_revoke_url, follow_redirects=True)
            assert res.status_code == 405



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Alembic upgrade tests."""

import pytest
from invenio_db import db


def test_alembic(app):
    """Test alembic recipes."""
    ext = app.extensions["invenio-db"]

    with app.app_context():
        if db.engine.name == "sqlite":
            raise pytest.skip("Upgrades are not supported on SQLite.")

        assert not ext.alembic.compare_metadata()
        db.drop_all()
        ext.alembic.upgrade()

        assert not ext.alembic.compare_metadata()
        ext.alembic.downgrade(target="96e796392533")
        ext.alembic.upgrade()

        assert not ext.alembic.compare_metadata()
        ext.alembic.downgrade(target="96e796392533")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import pytest
from flask import Flask

from invenio_oauth2server import InvenioOAuth2Server, InvenioOAuth2ServerREST
from invenio_oauth2server.ext import verify_oauth_token_and_set_current_user


def test_version():
    """Test version import."""
    from invenio_oauth2server import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioOAuth2Server(app)
    assert "invenio-oauth2server" in app.extensions
    assert ext.app is app

    app = Flask("testapp")
    ext = InvenioOAuth2Server()
    assert "invenio-oauth2server" not in app.extensions
    state = ext.init_app(app)
    assert "invenio-oauth2server" in app.extensions
    assert state.app is app


def test_init_rest():
    """Test REST extension initialization."""
    app = Flask("testapp")
    ext = InvenioOAuth2ServerREST(app)
    assert verify_oauth_token_and_set_current_user in app.before_request_funcs[None]

    app = Flask("testapp")
    ext = InvenioOAuth2ServerREST()
    assert verify_oauth_token_and_set_current_user not in app.before_request_funcs.get(
        None, []
    )
    ext.init_app(app)
    assert verify_oauth_token_and_set_current_user in app.before_request_funcs[None]


def test_init_rest_with_oauthlib_monkeypatch():
    """Test REST OAuthlib monkeypatching."""
    app = Flask("testapp")

    from oauthlib.common import urlencoded

    assert "^" not in urlencoded
    old_urlencoded = set(urlencoded)

    app.config["OAUTH2SERVER_ALLOWED_URLENCODE_CHARACTERS"] = "^"

    with pytest.warns(RuntimeWarning):
        InvenioOAuth2ServerREST(app)
    assert verify_oauth_token_and_set_current_user in app.before_request_funcs[None]

    from oauthlib.common import urlencoded

    assert old_urlencoded != urlencoded
    assert "^" in urlencoded
